;==============================================================================;
;       boot3.asm                                                              ;
; Code by Coty.                                                                ;
;                                                                              ;
;     This demo is to load the second sector of a hard disk drive using POI    ;
; and CHS format. Latter I might make some code to load with LBA28 and maybe   ;
; even with FAT 32 support, who knows? Any, no BIOS interupts are used in this ;
; code. We don't need them anyway. :-)                                         ;
;======================================v=======================================;
	use16                          ; Tell FASM to generate 16bit code.
	org	0x7C00                 ; Tell FASM were the proc will be loaded.
	format binary as "img"         ; Tell FASM to make a Hard Disk Image.
;======================================^=======================================;
;   Set up some stuff.                                                         ;
;======================================v=======================================;
	push	cs                     ; Get our current location.
	push	0xB800                 ; Push video memory.
	pop	gs                     ; pop video memory into GS.
	pop	ds                     ; ... And our current location into DS.
;======================================^=======================================;
; Enter and set up protected mode.                                             ;
;======================================v=======================================;
	cli                            ; Ignore interupts.
	lgdt	[gdtd]                 ; Load the GDT discriptor.
	mov	eax, cr0               ; Get controll reg data into EAX.
	inc	ax                     ; Set the PE bit.
	mov	cr0, eax               ; Put modified data into cr0.
	jmp	0x8:_PMode             ; Jump into protected mode.
;--------------------------------------;
;  GDT data area.                      ;
;--------------------------------------;
gdtd:	dw GDT_END - GDT_START - 1     ; Calculate GDT size.
	dd GDT_START                   ; Point processer to the gdt.
                                       ;
GDT_START:                             ;
	dq 0                           ; Create a null entry.
	;------------------------------;
	dw 0xFFFF                      ; Limmit of how much RAM to access (4GB)
	dw 0x0000                      ; Base adress of RAM 0x00000000
	dw 0x9A00                      ; code read/exec
	dw 0x00C0                      ; granularity
	;------------------------------;
	dw 0xFFFF                      ; access 4gb...
	dw 0x0000                      ; base adress
	dw 0x9200                      ; code read/exec
	dw 0x00C0                      ; granularity
GDT_END:                               ;
;--------------------------------------;
; Protected mode entry point.          ;
;--------------------------------------;
	use32                          ;
_PMode:                                ;
	mov	ax, 0x10               ; Point segment regs to the code 
	mov	ss, ax                 ;    descriptor.
	mov	es, ax                 ; 
	mov	ds, ax                 ; 
	mov	gs, ax                 ; 
	mov	fs, ax                 ; 
	sti                            ; Re-enable interupts.
;--------------------------------------;
; Make an IDT table.                   ;
;--------------------------------------;
	xor	edi, edi	       ; IDT is placed at 0x00000000
	mov	esi, IDT_snip	       ; The IDT snip
	mov	cx, 512	               ; 2048 bytes / 4 (4 byte regs)
	lodsd			       ; get the first half of the IDT snip
	xchg	eax, ebx               ; 
	lodsd			       ; Now the second part.
.idtloop:                              ;
	xchg	eax, ebx	       ; Swap EAX and EBX, EAX holds the first
				       ; Half of the IDT seg, and EBX holds the
				       ; Second part.
	stosd			       ; Store one part of the IDT
	loop	.idtloop	       ; Do this until CX == 0
;--------------------------------------;
; Map the PIC controller.              ;
;--------------------------------------;
	mov	 al, 00010001b	       ; bit 1: Use ICW4
				       ; bit 5: Intalise PIC
	out    0x20, al                ;
	mov	 al, 32 	       ; Tell PIC 1 what ISR number IRQ 0 is at
	out    0x21, al                ;
	mov	 al, 40                ;
	out    0xA1, al 	       ; Tell PIC 2 what ISR number IRQ 8 is at
                                       ;
	mov	 al, 0x4	       ; Tell the first PIC what line to
				       ; casscade
	out    0x21, al                ;
	mov	 al, 0x2	       ; Tell the second PIC what line to
				       ; cascade
	out    0xA1, al                ;
	mov	 al, 1		       ; Operate in intel (x86)
	out    0x21, al                ;
	out    0xA1, al                ;
	mov	 al, 11111111b	       ; Disable all IRQ's.
	out    0x21, al                ;
	mov	 al, 10111111b         ; Enable ATA Hard disk interupt.
	out    0xA1, al                ;
;======================================^=======================================;
;   Load sectors from HDD!                                                     ;
;======================================v=======================================;
	mov     dx, 0x01F6             ; Point DX to the "Drive and head" port
        ;------------------------------+--------------------------------------;
	mov	al, 0x00               ; put the head number into AL.
	mov	bl, 0x00               ; put desired drive # into BL.
	shl	bl, 4                  ; Shift bit to the 4th bit.
	or	bl, al                 ; Add resutls.
	mov	al, 0xA0               ; 0xA0 = master drive; 0xB0 = Slave.
	or	al, bl                 ; Get final results.
        ;------------------------------+--------------------------------------;
	out     dx, al                 ; Send drive and head to the DAH reg.
        ;******************************;
	mov     dx, 0x01f2             ; Point DX to the "Hard Disk Sector 
                                       ;    Count register".
	mov     al, 1                  ; Put number of sectors to read in AL(1)
	out     dx, al                 ; Send sector count to the HDSC reg...
        ;******************************;
	mov     dx, 0x01F3             ; Point DX to the "Sector Number" reg.
	mov     al, 2                  ; Toss the starting sector we wish to
                                       ;   Read into AL.
	out     dx, al                 ; Send desired sector the the HDC.
        ;******************************;
	mov     dx, 0x01f4             ; Point DX to the "Cylinder Low Port"
	mov     al, 0x00               ; Point AL to Cylender 0x00
	out     dx, al                 ; Send Cylender select low to CLP.
        ;******************************;
	mov     dx, 0x01F5             ; Point DX to "Cylinder High Port"
	out     dx, al                 ; Send 0x00 to the CHP.
        ;******************************;
	mov     dx, 0x01F7             ; Point EDX to the HDC command reg.
	mov     al, 0x20               ; Put command 0x20 "Read multiple sectors 
                                       ;    with retry" into AL.
	out     dx, al                 ; Send command to the HDC command reg.
;--------------------------------------;
; Wait for the controller to be ready. ;
;--------------------------------------+--------------------------------------;
; As HDC read from the HDD it will store all read data in register 0x01F0, we ;
; Can then read each byte/word/dword from this reg and store it into memory,  ;
; but first we must wait for HDD and HDC to finish loading our data.          ;
;--------------------------------------+--------------------------------------;
.loop:                                 ;
	in      al, dx                 ; Read from port 0x01F7
	test    al, 00001000b          ; Check bit 3, to see if HDC is ready!
                                       ;   Check "wait_HDC" function for more
                                       ;   information regarding these bits.
	jz      .loop                  ; If HDC != ready, do a loop!
;--------------------------------------;
; Read data into memory.               ;
;--------------------------------------;
	cli                            ; Disable interupts as recomended.

	mov	edi, 0x00010000            ; Point DI to 0x0000, were we will load
                                       ;   our kernel.
	mov	edx, 0x01F0	       ; Point DX to the Hard disk data reg.
	mov	ecx, 256               ; read 512 bytes. (256 words.)
                                       ;
	rep	insw                   ; Read word from port 0x01F0 to [ES:DI]
                                       ;   decrementing CX until it is empty.
        sti                            ; Enable Interupts.
        nop                            ;
	jmp	0x00010000             ; Jump to our kernel at 0x0001:0x0000
;======================================^========================================;
; Data area.                                                                    ;
;======================================v========================================;
idtr:	dw 2048 		       ; Size of IDT
	dd 0x00000000		       ; Location of IDT.
IDT_snip:                              ;
	dw dummy_int		       ; Interupt location.
	dw 0x0008		       ; 0x0008 = offset of the code descriptor
				       ; in the GDT.
	dw 0x8E00		       ; This is required.
	dw 0			       ; This needs to be a zero.
                                       ;
dummy_int:                             ; 
	mov	al, 0x20               ; Send acknollage interupt command.
	out	0x20, al               ; Send command to the first PIC...
	out	0xA0, al               ; ... And the second PIC.
	iret                           ; Return to exicution before the interupt.
                                       ;
times (510-($-$$)) db 0                ; Make our boot sector.
        dw 0xAA55                      ; Add BIOS boot signature.
;==============================================================================;
; Our kernel.                                                                  ;
;======================================v=======================================;
	org	0x00010000             ; Tell FASM were loaded at 0x0001:0x0000
_kernel:                               ;
	mov	edi, 0x000B8000        ; Point EDI to video memory.
	mov	eax, 0x00000000        ; Empty eax.
	mov	ecx, 80*25/2           ; Tell ECX how many colitions there are.
	rep	stosb                  ; Clear them.
                                       ; 
	mov	esi, msg_hello         ; Point ESI to the msg string.
	mov	edi, 0x000B8000        ; point EDI to video memory.
	mov	 ah, 0x0F              ; White on black.
.loop1:                                ;
	lodsb                          ; Get byte from message string.
	cmp	al, 0x00               ; Is byte 0?
	je	.exit                  ; If so exit loop.
	stosw                          ; Print byte.
	jmp	.loop1                 ; Do again until byte is 0.
.exit:                                 ;
                                       ;
	jmp	$                      ; Stop...
                                       ;
msg_hello db "Xeron-OS v0.05", 0       ;
                                       ;
times (2097152-($-$$)) db 0            ; Make a 2 MB hard disk image.
