;==============================================================================;
;       pr0.asm                                                                ;
; Code (C) 2011 Coty R Miller.                                                 ;
;                                                                              ;
;     This demo is to load "kernel.sys" from a XFS (Xeron File System) v1      ;
; formated hard disk in support of LBA 28 and with some mods LBA 48.           ;
;======================================v=======================================;
	use16                          ; Tell FASM to generate 16bit code.
	org	0x7C00                 ; Tell FASM were the proc will be loaded.
	format binary as "img"         ; Tell FASM to make a Hard Disk Image.
;======================================^=======================================;
;   XFS (Xeron File System) data.                                              ;
;======================================v=======================================;
	jmp	short .start           ; Jump short REQUIRED to code start.
	db	"X-FS"                 ; Declare this file system is XFS
	db	1                      ; XFS version
	db	"DISK LABLE  "         ; Disk label.
;======================================^=======================================;
;   Set up some stuff.                                                         ;
;======================================v=======================================;
.start:
	push	cs                     ; Get our current location.
	push	0xB800                 ; Push video memory.
	pop	gs                     ; pop video memory into GS.
	pop	ds                     ; ... And our current location into DS.
;======================================^=======================================;
; Enter and set up protected mode.                                             ;
;======================================v=======================================;
	cli                            ; Ignore interupts.
	lgdt	[gdtd]                 ; Load the GDT discriptor.
	mov	eax, cr0               ; Get controll reg data into EAX.
	inc	ax                     ; Set the PE bit.
	mov	cr0, eax               ; Put modified data into cr0.
	jmp	0x8:_PMode             ; Jump into protected mode.
;--------------------------------------;
;  GDT data area.                      ;
;--------------------------------------;
gdtd:	dw GDT_END - GDT_START - 1     ; Calculate GDT size.
	dd GDT_START                   ; Point processer to the gdt.
                                       ;
GDT_START:                             ;
	dq 0                           ; Create a null entry.
	;------------------------------;
	dw 0xFFFF                      ; Limmit of how much RAM to access (4GB)
	dw 0x0000                      ; Base adress of RAM 0x00000000
	dw 0x9A00                      ; code read/exec
	dw 0x00C0                      ; granularity
	;------------------------------;
	dw 0xFFFF                      ; access 4gb...
	dw 0x0000                      ; base adress
	dw 0x9200                      ; code read/exec
	dw 0x00C0                      ; granularity
GDT_END:                               ;
;--------------------------------------;
; Protected mode entry point.          ;
;--------------------------------------;
	use32                          ;
_PMode:                                ;
	mov	ax, 0x10               ; Point segment regs to the code 
	mov	ss, ax                 ;    descriptor.
	mov	es, ax                 ; 
	mov	ds, ax                 ; 
	mov	gs, ax                 ; 
	mov	fs, ax                 ; 
	sti                            ; Re-enable interupts.
;--------------------------------------;
; Make an IDT table.                   ;
;--------------------------------------;
	xor	edi, edi	       ; IDT is placed at 0x00000000
	mov	esi, IDT_snip	       ; The IDT snip
	mov	cx, 512	               ; 2048 bytes / 4 (4 byte regs)
	lodsd			       ; get the first half of the IDT snip
	xchg	eax, ebx               ; 
	lodsd			       ; Now the second part.
.idtloop:                              ;
	xchg	eax, ebx	       ; Swap EAX and EBX, EAX holds the first
				       ; Half of the IDT seg, and EBX holds the
				       ; Second part.
	stosd			       ; Store one part of the IDT
	loop	.idtloop	       ; Do this until CX == 0
;--------------------------------------;
; Map the PIC controller.              ;
;--------------------------------------;
	mov	 al, 00010001b	       ; bit 1: Use ICW4
				       ; bit 5: Intalise PIC
	out    0x20, al                ;
	mov	 al, 32 	       ; Tell PIC 1 what ISR number IRQ 0 is at
	out    0x21, al                ;
	mov	 al, 40                ;
	out    0xA1, al 	       ; Tell PIC 2 what ISR number IRQ 8 is at
                                       ;
	mov	 al, 0x4	       ; Tell the first PIC what line to
				       ; casscade
	out    0x21, al                ;
	mov	 al, 0x2	       ; Tell the second PIC what line to
				       ; cascade
	out    0xA1, al                ;
	mov	 al, 1		       ; Operate in intel (x86)
	out    0x21, al                ;
	out    0xA1, al                ;
	mov	 al, 11111111b	       ; Disable all IRQ's.
	out    0x21, al                ;
	mov	 al, 10111111b         ; Enable ATA Hard disk interupt.
	out    0xA1, al                ;
;--------------------------------------^----------------------------------------;
; Load and interpret XFS (Xeron File System)                                    ;
;--------------------------------------v----------------------------------------;
	mov	eax, 2                 ; Point EAX at the second sector.
	mov	ebx, 1                 ; Tell EBX that we want to load 1 sector.
	mov	edi, 0x7E00            ; Point EDI to were we want to load it.
	call	Load_LBA28             ; Load it using LBA 28...
                                       ;
		mov	edi, kname     ; Point EDI to the kernel name.
		mov	esi, entry0    ; Point ESI to the entry table.
		mov	ecx, knlngh    ; Tell ECX how long the kernel name is.
		xor	edx, edx       ; Empty EDX...
		xor	eax, eax       ; ... and eax.
	xfs_loop:                      ;
		REPE	CMPSB          ; Compare the strings until they dont match
		cmp	 cl, 0         ; ... or ECX == 0
		jne	xfs_nogo       ; IF ecx != 0 THEN GOTO xfs_nogo
                                       ;
		sub	esi, knlngh    ; Subtract kernel name size from ESI
		add	esi, 27        ; add 27 to ESI to point at entry data.
		lodsd                  ; Load dword of sector location.
		push	eax            ; Save it sector onto stack.
		xor	eax, eax       ; empty EAX.
		lodsb                  ; Load the sector length byte.
		mov	ebx, eax       ; Put sector lengh into ebx
		pop	eax            ; Restore EAX.
		jmp	xfs_done       ; Our job here is done!
	xfs_nogo:                      ; 
		cmp	 dl, 0x0E      ; Is EDX 0xE? If so, we have checked all entries
		je	xfs_notfound   ;  so file not found :(
		inc	edx            ; If not, add 1 to EDX.
                                       ;
		push	eax            ; Save EAX.
		mov	eax, knlngh    ; Put kernel file length name into EAX
		sub	eax, ecx       ; Subtract ECX from EAX to discover were ESI is.
		mov	ecx, eax       ; Put info into ECX...
		pop	eax            ; Restore EAX.
                                       ;
		mov	edi, kname     ; Point EDI back at the kernel name. 
		sub	esi, ecx       ; Subtract ECX from ESI.
		add	esi, 32        ; Move esi to the next table entry
		mov	ecx, knlngh    ; Put kernel length name size into eax.
		jmp	xfs_loop       ; Search next entry...
                                       ;
xfs_done:                              ;
	mov 	edi, 0x00010000        ;
	call	Load_LBA28             ;
                                       ;
	jmp	0x00010000             ; Jump to our kernel at 0x0001:0x0000
                                       ;
xfs_notfound:                          ;
	mov	dword[0xB8000], ": ( " ; Print a frowny face because the kernel
                                       ;   wasn't found :(

	hlt                            ; Halt the system...
	Jmp	$-1                    ; If we make it here, go back to the halt
;======================================^=======================================;
;   Load sectors from HDD using LBA!                                           ;
;      EAX = sector to load                                                    ;
;      EBX = sectors to load                                                   ;
;      EDI = Place to load data.                                               ; 
;======================================v=======================================;
Load_LBA28:                            ;
	push	eax                    ; Store sector number
	mov 	edx, 0x01F2            ; 0x01F2 - Sector count Port
	mov 	 al, bl                ; Read sector count in BL
	out	 dx, al                ;
                                       ;
	pop 	eax                    ; Restore sectore number
	inc	 dx		       ; 0x01F3 - LBA Low Port
	out	 dx, al                ;
                                       ;
	inc 	edx                    ; 0x01F4 - LBA Mid Port
	shr 	eax, 8                 ; get 2nd location byte
	out	 dx, al                ;
                                       ;
	inc 	edx                    ; 0x01F5 - LBA High Port
	shr 	eax, 8                 ;
	out	 dx, al                ;
                                       ;
	inc 	edx                    ; 0x01F6 - Device Port. 
	mov	 al, 0100000b          ; bit:
                                       ;    6 == 1: Use LBA adressing
                                       ;    4 == 0: Master
                                       ;         1: Slave
                                       ;  3-0 ==  : 0 for LBA "Extra high"
                                       ;
	out	 dx, al                ;
                                       ;
	inc 	edx                    ; 0x01F7 - Command Port
	mov	 al, 0x20              ; Read sectors command ((0x24 if LBA48))
	out	 dx, al                ;
;--------------------------------------;
; Wait for the controller to be ready. ;
;--------------------------------------+--------------------------------------;
; As HDC read from the HDD it will store all read data in register 0x01F0, we ;
; Can then read each byte/word/dword from this reg and store it into memory,  ;
; but first we must wait for HDD and HDC to finish loading our data.          ;
;--------------------------------------+--------------------------------------;
.loop:                                 ;
	in      al, dx                 ; Read from port 0x01F7
	test    al, 00001000b          ; Check bit 3, to see if HDC is ready!
                                       ;   Check "wait_HDC" function for more
                                       ;   information regarding these bits.
	jz      .loop                  ; If HDC != ready, do a loop!
;--------------------------------------;
; Read data into memory.               ;
;--------------------------------------;
	cli                            ; Disable interupts as recomended.
	inc	ebx                    ; Increase sectors to read by 1.
	imul	ecx, ebx, 256          ; sectors * 256 = ECX, "words to read"
                                       ;
	mov	edx, 0x01F0	       ; Point DX to the Hard disk data reg.
	mov	ecx, 256               ; read 512 bytes. (256 words.)
                                       ;
	rep	insw                   ; Read word from port 0x01F0 to [ES:DI]
                                       ;   decrementing CX until it is empty.
        sti                            ; Enable Interupts.
	ret                            ; Return to who called us.
;======================================^========================================;
; Data area.                                                                    ;
;======================================v========================================;
idtr:	dw 2048 		       ; Size of IDT
	dd 0x00000000		       ; Location of IDT.
IDT_snip:                              ;
	dw dummy_int		       ; Interupt location.
	dw 0x0008		       ; 0x0008 = offset of the code descriptor
				       ; in the GDT.
	dw 0x8E00		       ; This is required.
	dw 0			       ; This needs to be a zero.
                                       ;
dummy_int:                             ; 
	mov	al, 0x20               ; Send acknollage interupt command.
	out	0x20, al               ; Send command to the first PIC...
	out	0xA0, al               ; ... And the second PIC.
	iret                           ; Return to exicution before the interupt.
                                       ;
kname: db "kernel.sys"                 ; Kernel file name.
knlngh = $ - kname                     ; Kernel file name lengh
                                       ;
times (510-($-$$)) db 0                ; Make our boot sector.
        dw 0xAA55                      ; Add BIOS boot signature.
;======================================^=======================================;
;   XFS (Xeron File System) v1 table.                                          ;
;======================================v=======================================;
entyta dw 0000000000000001b            ;  Ocupied entries contain a 1.
                                       ;
entry0 db "kernel.sys                 "; File name.
       dd 0x00000003                   ; File sector location.
       db 1                            ; How many sectors long the file is -1.
entry1 db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entry2 db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entry3 db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entry4 db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entry5 db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entry6 db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entry7 db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entry8 db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entry9 db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entryA db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entryB db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entryC db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entryD db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
entryE db "                           "; File name.
       dd 0x00000000                   ; File sector location.
       db 0                            ; How many sectors long the file is -1
                                       ;
times (1024-($-$$)) db 0               ; Table entry exactly 512 bytes.
;==============================================================================;
; Our kernel.                                                                  ;
;======================================v=======================================;
	org	0x00010000             ; Tell FASM were loaded at 0x0001:0x0000
_kernel:                               ;
	mov	edi, 0x000B8000        ; Point EDI to video memory.
	mov	eax, 0x00000000        ; Empty eax.
	mov	ecx, 80*25/2           ; Tell ECX how many colitions there are.
	rep	stosb                  ; Clear them.
                                       ; 
	mov	esi, msg_hello         ; Point ESI to the msg string.
	mov	edi, 0x000B8000        ; point EDI to video memory.
	mov	 ah, 0x0F              ; White on black.
.loop1:                                ;
	lodsb                          ; Get byte from message string.
	cmp	al, 0x00               ; Is byte 0?
	je	.exit                  ; If so exit loop.
	stosw                          ; Print byte.
	jmp	.loop1                 ; Do again until byte is 0.
.exit:                                 ;
                                       ;
	jmp	$                      ; Stop...
                                       ;
msg_hello db "Xeron-OS v0.05", 0       ;
                                       ;
times (2097152-($-$$)) db 0            ; Make a 2 MB hard disk image.
