;=============================================================================;
; Helium-OS (TM) 2011 Coty Miller                                             ;
; Code Copyright by Coty Miller 2010                                          ;
; compile with "fasm boot.asm boot.bin"                                       ;
; Email: compcoty@gmail.com                                                   ;
;-----------------------------------------------------------------------------;
; Helium-OS is a 32-bit p-mode multi tasking Operating system smaller than    ;
; 512 bytes... It would probibly run best with 4MB of RAM, you can use more   ;
; but this OS does not allocate it. I would say the bare minimum requirements ;
; would be an intel based 32bit processer (of course) and 32kb of RAM         ;
; assuming that a fault is not issued if the OS allocates more RAM than you   ;
; have...                                                                     ;
;                                                                             ;
; This OS is made to boot from a floppy disk, but with a few adjustments you  ;
; could probibly get it to boot from a CD-ROM... It probibly good to note     ;
; that I tried to cram this OS into a very small space. So it could be way    ;
; more readible. :-)                                                          ;
;                                                                             ;
; !-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-! ;
;                                !  DISCLAMER !                               ;
; YOU MAY NOT HOLD ME RESPONSIBLE FOR ANY DATA LOSS/SMOKE YOU COMPUTER MAKES  ;
; AND ALL OF THOSE ABOVE THAT INVOLVE CRASHING YOUR PC AND STUFF! USE AT YOUR ;
; OWN RISK!                                                                   ;
;-----------------------------------------------------------------------------;
; Thanks to:                                                                  ;
;      Dex, f2, qandrew, those who contributed on OSDev wiki, and those on    ;
;      the FASM board and DexOS comunity that helped.                         ;
;=============================================================================;
; Version 0.01.6
format binary as "img"
	ORG	0x7C00
	use16
	cli			       ; Disable interrupts
	xor	 bx, bx
	mov	 ds, bx 	       ; DS == 0
	mov	 es, bx 	       ; ES == 0
	;'''''''''''''''''''''''''''''';
	; Set 80x25 text mode.         ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	mov	ax, 0x1112	       ; Set 80x50 mode
	int   0x10		       ; Do it!
;--------------------------------------;
; ********** Set up paging *********** ;
;--------------------------------------;
; I'll map 4MB of memory...
; use 4kb pages...
	mov	 di, 0x1000	       ; Paging starts at 0x1000 (at the PML4)
	mov	cr3, edi	       ; place the location of PML4 into Cr3
	push	edi		       ; [!] put 0x00001000 on the stack. [!]
	;'''''''''''''''''''''''''''''';
	; Create the page directories. ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	xor	eax, eax
	mov	 al, 0x2	       ; NOT present
	mov	 cx, 1024	       ; 1024 entries... 4MB
	push	 cx		       ; {!} put 1024 onto the stack. {!}
	rep	stosd		       ; Clear all 1024 entries
	;'''''''''''''''''''''''''''''';
	; Create 1 page table, for 4MB ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	pop	 cx		       ; {!} Get 1024 from the stack. {!}
	mov	 di, 0x2000
.loop:
	or	 al, 3		       ; <<< eax
	stosd
	add	eax, 4096
	loop	.loop
	;'''''''''''''''''''''''''''''';
	; Point page DIR to page table ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	pop	edi		       ; [!] Get 0x00001000 from the stack [!]
	mov	eax, 0x00002003
	stosd
; page_directory   = 0x1000
; first_page_table = 0x2000
;--------------------------------------;
; ********** Enter P-mode ************ ;
;--------------------------------------;
	LGDT	[gdtr]		       ; Load the GDT
	mov	eax, cr0
	or	eax, 0x80000001        ; Enable PE bit, and paging bit.
	mov	cr0, eax
	jmp	0x8:_Kernel	       ; CS point to code sel + jmp into P-mode
;-----------------------------------------------------------------------------;
; *********************************** GDT *********************************** ;
;-----------------------------------------------------------------------------;
gdtr:	dw GDT_end - GDT_start - 1     ; GDT size - 1  So that should be 23.
	dd GDT_start		       ; Were the GDT starts.
GDT_start:
	dq 0x0000000000000000	       ; Null descriptor.
	dw 0x03FF		       ; 4MB limmit  1023 (1024-1)*4096 = 4MB)
	dw 0x0000		       ; Base adress = 0x0000
	dw 0x9A00		       ; code read/exec
	dw 0x00C0		       ; granularity

	dw 0x03FF		       ; 4MB limmit  1023 ((1024-1)*4096=2MB)
	dw 0x0000		       ; Base adress = 0x0000
	dw 0x9200		       ; code read/exec
	dw 0x00C0		       ; granularity
GDT_end:
;-----------------------------------------------------------------------------;
; *********************************** IDT *********************************** ;
;-----------------------------------------------------------------------------;
	use32
idtr:	dw 2048 		       ; Size of IDT
	dd 0x00000000		       ; Location of IDT.
IDT_snip:
	dw dummy_int		       ; Interupt location.
	dw 0x0008		       ; 0x0008 = offset of the code descriptor
				       ; in the GDT.
	dw 0x8E00		       ; This is required.
	dw 0			       ; This needs to be a zero.
dummy_int:
	mov	 ax, 0x0058	       ; AH == 0x00, print Char
				       ; AL == 'X' , letter to print.
	int	0x70
	cli
	hlt
;=============================================================================;
; ******************************* Main setup ******************************** ;
;=============================================================================;
_Kernel:
	mov	 ax, 0x0010	       ; Move the data segment into AX then
	mov	 ss, ax 	       ; into SS, DS, and ES.
	mov	 ds, ax
	mov	 es, ax
	; GS and FS will remain as they
	; are, they are not important to
	; set.
	; Note that since were in P-mode
	; we must not worry with segmentation.
;--------------------------------------;
; ********* Clear the screen ********* ;
;--------------------------------------;
	mov	edi, 0x000B8000        ; Text mode video memory.
	mov	eax, 0x07000700
	mov	 cx, 2000	       ; 80*50 / 2
	rep	stosd		       ; write eax to the video 1000 times.
;--------------------------------------;
; ********* Install The IDT ********** ;
;--------------------------------------;
	xor	edi, edi	       ; IDT is placed at 0x00000000
	mov	esi, IDT_snip	       ; The IDT snip
	mov	 cx, 512	       ; 2048 bytes / 4 (4 byte regs)
	lodsd			       ; get the first half of the IDT snip
	xchg	eax, ebx
	lodsd			       ; Now the second part.
.idtloop:
	xchg	eax, ebx	       ; Swap EAX and EBX, EAX holds the first
				       ; Half of the IDT seg, and EBX holds the
				       ; Second part.
	stosd			       ; Store one part of the IDT
	loop	.idtloop	       ; Do this until CX == 0
;--------------------------------------;
; ********** Program the PIT ********* ;
; Set PIT to 100Hz (0x2E 94)           ;
;--------------------------------------;
	;'''''''''''''''''''''''''''''';
	; Set low byte.                ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	mov	 al, 00010110b	       ; control register
	out    0x43, al 
	mov	 al, 0x2E	       ; LSB of clock divisor
	out    0x40, al
	;'''''''''''''''''''''''''''''';
	; Set high byte.               ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	mov	 al, 00100110b	       ; Contol register.
	out    0x43, al
	mov	 al, 0x9A	       ; MSB only clock divisor
	out    0x40, al 	       ; Wait for PIT to finish...
;--------------------------------------;
; ************ Map the PIC *********** ;
;--------------------------------------;
	mov	 al, 00010001b	       ; bit 1: Use ICW4
				       ; bit 5: Intalise PIC
	out    0x20, al
	mov	 al, 32 	       ; Tell PIC 1 what ISR number IRQ 0 is at
	out    0x21, al
	mov	 al, 40
	out    0xA1, al 	       ; Tell PIC 2 what ISR number IRQ 8 is at

	mov	 al, 0x4	       ; Tell the first PIC what line to
				       ; casscade
	out    0x21, al
	mov	 al, 0x2	       ; Tell the second PIC what line to
				       ; cascade
	out    0xA1, al
	mov	 al, 1		       ; Operate in intel (x86)
	out    0x21, al
	out    0xA1, al
	mov	 al, 11111100b	       ; Disable all IRQ's except for the kbd
				       ; and the timmer.
	out    0x21, al
	mov	 al, 0xFF	       ; No interupts enabled on PIC 2.
	out    0xA1, al
;--------------------------------------;
; ******** Open the A-20 gate ******** ;
;--------------------------------------;
	mov	 al, 0xAD	       ; Disable the system keyboard.
	out    0x64, al
	call	wait_for_8042	       ; Wait until the beyboard is ready.
	mov	 al, 0xD1	       ; Instruct the keyboard to write to pin2
	out    0x64, al

	call	wait_for_8042	       ; Wait for the system keyboard.
	mov	 al, 0xDF	       ; Enable the A-20 address line.
	out    0x60, al

	mov	 al, 0xAE	       ; Re-enable the keyboard.
	out    0x64, al
	call	wait_for_8042	       ; Wait for it.
;--------------------------------------;
; ********* Hook needed IRQs ********* ;
;--------------------------------------;
	xor	edi, edi
	mov	 di, 32*8	       ; Timer IRQ location.
	mov	 ax, Timer_handle      ; New Timmer IRQ handle
	stosw
	mov	 di, 33*8	       ; Keyboard IRQ location.
	mov	 ax, kbd_handle        ; New Timmer IRQ handle
	stosw
;--------------------------------------;
; ***** Hook software Interupts. ***** ;
;--------------------------------------;
	mov	 di, 0x70*8	       ; Int 0x70 handle location.
	mov	 ax, _int_0x70	       ; New Timmer IRQ handle
	stosw
;--------------------------------------;
; ******* Install multitasking ******* ;
;--------------------------------------;
	mov	esp, stack1_base       ; Declare task 1 stack. :-)
	;'''''''''''''''''''''''''''''';
	; Put stuff onto the stack for ;
	; IRET/IRETD.                  ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	sti			       ; Enable interups (Delays for 1 instruction
	mov	eax, Task1	       ; point eax to task1 location.
	pushf			       ; Push Eflags
	cli			       ; Re-enable interupts.
	push	 cs		       ; Push CS
	push	eax		       ; Push EIP (well, it will be EIP :-).

	pushad			       ; Save all other regs.
	;'''''''''''''''''''''''''''''';
	; Swap to stack 0. :-)         ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	mov	 di, _esp	       ; <<< EDI
	mov	esi, edi
	lodsd
	xchg	eax, esp
	stosd
	;'''''''''''''''''''''''''''''';
	; Goto task 0.  .\m/           ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	sti
	jmp	Task0		       ; Goto Task 0.
;=============================================================================;
; Functions go here.                                                          ;
;=============================================================================;
wait_for_8042:
	jmp	$+2		       ; wait for second

	in	 al, 0x64	       ; read from KBD
	test	 al, 2		       ; is it 2?
	jnz	wait_for_8042	       ; If not wait...
	ret
;============================================================================;
; ***************************** Timmer handler. **************************** ;
; This is used for task switching. :-)                                       ;
;============================================================================;
Timer_handle:
; The CPU has pushed the following onto the stack.
; eflags, cs, eip.
	cli
	pushad			      ; Save regs.
	;''''''''''''''''''''''''''''';
	; Swap ESP's stack locations. ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	mov	esi, _esp	      ; ESP save spot into ESI.
	mov	edi, esi	      ; Put it into EDI as well.
	lodsd			      ; Load double word
	xchg	eax, esp	      ; exchange EAX and ESP.
	stosd			      ; Save old ESP.
	;''''''''''''''''''''''''''''';
	; Prepare return to other tsk ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
_return:
	mov	 al, 0x20	      ; Tell PICs that the IRQ was handled.
	out    0xA0, al 	      ; PIC 1
	out    0x20, al 	      ; PIC 2
	popad			      ; Restore regs.
	; Normaly you would want to use STI, but there is no need
	; As IRET pops the eflags reg of the stack, there is a 99%
	; chance that interupts are enabled :-)
	iretd
_esp:	dd 0x00007100  ; Extra stack pointer save location.
stack0_base = 0x00007100
stack1_base = 0x00007200
;============================================================================;
; **************************** Keyboard handler ***************************  ;
;============================================================================;
kbd_handle:
	cli
	pushad
	in	al, 0x60	       ; Read keypress from keyboard.
	cmp	al, 1		       ; Did mr rude press ESC?
	jne	_return 	       ; If so not do not re-stat!
restart:			       ; If so restart.
	mov	 al, 0xFE	       ; 0xFE sent to keyboard controller
.loop:
	out    0x64, al 	       ; trigers a re-start.
	jmp	.loop		       ; Continue untill restart...
;============================================================================;
; ****************************** Interupt 0x70 ***************************** ;
;============================================================================;
_int_0x70:
	cli
	pushad
	cmp	 ah, 1		       ; Is AH == 1?
	je	get_kbd 	       ; If so get k_press.
	;'''''''''''''''''''''''''''''';
	; INT 0x70 AH = 0, print char  ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	push	eax
	mov	esi, _cur	       ; Get current video mem location.
	lodsd
	mov	edi, eax
	pop	eax
	mov	 ah, 0x02	       ; Color of the desired text...
	stosw			       ; Put charictar
	cmp	 di, 0x9F40	       ; Is edi pointing at 80x50?
	jne	.continue	       ; If not continue on.
	;'''''''''''''''''''''''''''''';
	; Scroll screen down 1 line.   ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
	mov	esi, 0x000B8050        ; point ESI to vid mem + 80
	mov	 di, 0x8000	       ; EDI == 0x000B8000
	mov	ecx, 80*49	       ; ECX == how many screen cols there are
	rep	movsw		       ; Copy all text up one line.
; Now clear bottom line...
	mov	 ax, 0x0700	       ; Colour = gray on black
	mov	 di, 0x9EA0	       ; EDI =0x000B9EA0 (begining of last line
	push	edi		       ; [!] Save EDI [!]
	mov	 cx, 80 	       ; Clear 80 cols
	rep	 stosw		       ; Do it!
	pop	edi		       ; [!] Restore EDI [!]
	;'''''''''''''''''''''''''''''';
	; Save cursor location.        ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
.continue:
	mov	eax, _cur	       ; Load cursor location into EAX.
	xchg	edi, eax	       ; Put cursor location into EAX and
	stosd			       ; _cur location into EDI.
	jmp	_int_0x70_exit
_cur:	dd 0x000B8000
	;'''''''''''''''''''''''''''''';
	; INT 0x70 AH = 1, get k press ;
	;,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,;
get_kbd:
	cmp	byte[kbdpress], 1      ; Is kbd key press?
	jne	_int_0x70_exit	       ; If not return.
	in	 al, 0x60	       ; Read key press from kbd controller
	mov	byte[kbdpress], 0      ; Empty kbd status... (no key press)
_int_0x70_exit:
	popad			       ; Pop off all interupts
	iretd			       ; Return.
;=============================================================================;
; Task A and B go here...                                                     ;
;=============================================================================;
Task0:
	Mov	 ah, 0		       ; INT 0x70, AH = 0, Print Char
	Mov	 al, '0'	       ; Print a 0
.loop:
	Int    0x70		       ; Print
	hlt			       ; print only 1 char each time the task
				       ; is returned to...
	Jmp	.loop
end_0:

Task1:
	Mov	 ah, 0		       ; INT 0x70, AH = 0, Print Char
	Mov	 al, '1'	       ; Print a 1
.loop:
	Int    0x70		       ; Print
	hlt			       ; print only 1 char each time the task
				       ; is returned to...
	Jmp	.loop
end_1:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	times (510-($-$$)) db 0
	dw	0xAA55		       ; These two bytes verify that the sector
				       ; was loaded sucsesfully.
;=============================================================================;
; Data...                                                                     ;
; This data is set at 0x7E00, right after the boot sectors, this is were we   ;
; can save a few values! :-)                                                  ;
;=============================================================================;
kbdpress:			       ; 1 byte.

	times (368640-($-$$)) db 0     ; 360kb floppy.
